import { render } from 'lit-html';
export { html } from 'lit-html';

const tryParseInt = (value) => {
    return (parseInt(value) == value && parseFloat(value) !== NaN) ? parseInt(value) : value;
};
const getSetProps = (target) => {
    const proto = target.prototype;
    return Object.getOwnPropertyNames(proto)
        .filter(s => (typeof proto[s] != 'function'))
        .filter(key => !/^(|length|arguments|caller)$/.test(key))
        .reduce((acc, cur, i) => {
        acc[cur] = null;
        return acc;
    }, {});
};
const initProps = (target) => {
    const props = target.constructor.props || {};
    const decorators = target.constructor.propDecorators;
    for (const prop of Object.keys(decorators)) {
        decorators[prop] = target[prop];
    }
    return Object.assign({}, props, decorators);
};
const autoBind = (element) => {
    const proto = element.constructor.prototype;
    const propertyNames = Object.getOwnPropertyNames(proto)
        .filter(s => (typeof element[s] == 'function'))
        .filter(key => !/^(prototype|name|constructor|render|connectedCallback|attributeChangedCallback)$/.test(key));
    for (const prop of propertyNames) {
        element[prop] = element[prop].bind(element);
    }
};
const toKebabCase = (str) => {
    return str
        .replace(/([a-z])([A-Z])/g, '$1-$2')
        .replace(/[\s_]+/g, '-')
        .toLowerCase();
};
const tryParseValue = (value) => {
    const pattern = /^([Tt][Rr][Uu][Ee]|[Ff][Aa][Ll][Ss][Ee])$/g;
    if (value) {
        if (typeof value == 'boolean') {
            return (/^(true|false|^$)$/.test(value.toString()));
        }
        if (typeof value == 'string') {
            const _value = value.match(pattern);
            if (_value && Array.isArray(_value) && /^(true|false|^$)$/.test(_value[0])) {
                return JSON.parse(_value[0]);
            }
        }
    }
    return tryParseInt(value);
};

const renderTemplate = (rootElement) => {
    
    render(rootElement.render(), rootElement.shadowRoot);
};
class LitCustomElement extends HTMLElement {
    constructor() {
        super();
        this.values = new Map();
        this.attachShadow({ mode: 'open' });
        autoBind(this);
    }
    static get observedAttributes() {
        this.fieldProps = Object.assign({}, getSetProps(this));
        return Object.keys(Object.assign({}, this.constructor.props, this.propDecorators, this.fieldProps))
            .map(prop => toKebabCase(prop));
    }
    attributeChangedCallback(name, oldValue, newValue) {
        oldValue = tryParseValue(oldValue), newValue = tryParseValue(newValue);
        if (oldValue !== newValue) {
            if (this[name] !== newValue) {
                this[name] = newValue;
            }
            renderTemplate(this);
        }
    }
    setAttribute(qualifiedName, value) {
        this.values.set(qualifiedName, tryParseValue(value));
        super.setAttribute(qualifiedName, this.values.get(qualifiedName));
    }
    propertyChanged(prop, attribute = true) {
        const descriptor = Object.getOwnPropertyDescriptor(this.constructor.prototype, prop);
        Object.defineProperty(this, prop, Object.assign({}, descriptor, { get() {
                return descriptor.get.call(this);
            },
            set(value) {
                descriptor.set.call(this, value);
                this.onPropertyChanged(prop, attribute);
            } }));
        if (attribute && this[prop]) {
            this.setAttribute(prop, this[prop]);
        }
    }
    onPropertyChanged(propName, attribute) {
        const propValue = tryParseValue(this[propName]);
        if (this.values.get(propName) !== propValue) {
            this.values.set(propName, propValue);
            if (attribute) {
                super.setAttribute(propName, propValue);
            }
            renderTemplate(this);
        }
    }
    connectedCallback() {
        const props = initProps(this);
        
        this.initPropAccessors &&
            
            this.initPropAccessors(Object.keys(this.constructor.fieldProps));
        for (const prop of Object.keys(props)) {
            const propName = toKebabCase(prop);
            Object.defineProperty(this, prop, {
                get() { return this.values.get(propName); },
                set(value) { this.setAttribute(propName, value); }
            });
            if (this.hasAttribute(propName)) {
                this.setAttribute(propName, this.getAttribute(propName));
            }
            else {
                this.setAttribute(propName, props[prop]);
            }
        }
        renderTemplate(this);
    }
}
LitCustomElement.propDecorators = {};
LitCustomElement.fieldProps = {};

const Prop = () => {
    return (target, propName) => {
        target.constructor.propDecorators[propName] = null;
    };
};

export { LitCustomElement, Prop, autoBind, getSetProps, initProps, toKebabCase, tryParseInt, tryParseValue };
